
.. DO NOT EDIT.
.. THIS FILE WAS AUTOMATICALLY GENERATED BY SPHINX-GALLERY.
.. TO MAKE CHANGES, EDIT THE SOURCE PYTHON FILE:
.. "auto_examples/anatomical/example_T2s_EPI.py"
.. LINE NUMBERS ARE GIVEN BELOW.

.. only:: html

    .. note::
        :class: sphx-glr-download-link-note

        :ref:`Go to the end <sphx_glr_download_auto_examples_anatomical_example_T2s_EPI.py>`
        to download the full example code.

.. rst-class:: sphx-glr-example-title

.. _sphx_glr_auto_examples_anatomical_example_T2s_EPI.py:


Compare Fourier Model and T2* Model for EPI
===========================================

This examples walks through the elementary components of SNAKE.

Here we proceed step by step and use the Python interface. A more integrated
alternative is to use the CLI ``snake-main``

.. GENERATED FROM PYTHON SOURCE LINES 14-22

.. code-block:: Python


    # Imports
    import numpy as np
    from snake.core.simulation import SimConfig, default_hardware, GreConfig
    from snake.core.phantom import Phantom
    from snake.core.smaps import get_smaps
    from snake.core.sampling import EPI3dAcquisitionSampler








.. GENERATED FROM PYTHON SOURCE LINES 23-36

.. code-block:: Python


    sim_conf = SimConfig(
        max_sim_time=6,
        seq=GreConfig(TR=100, TE=30, FA=3),
        hardware=default_hardware,
        fov_mm=(181, 217, 181),
        shape=(60, 72, 60),
    )
    sim_conf.hardware.n_coils = 8

    phantom = Phantom.from_brainweb(sub_id=4, sim_conf=sim_conf, tissue_file="tissue_7T")




.. rst-class:: sphx-glr-script-out

.. code-block:: pytb

    Traceback (most recent call last):
      File "/home/runner/work/snake-fmri/snake-fmri/examples/anatomical/example_T2s_EPI.py", line 33, in <module>
        phantom = Phantom.from_brainweb(sub_id=4, sim_conf=sim_conf, tissue_file="tissue_7T")
      File "/home/runner/work/snake-fmri/snake-fmri/src/snake/core/phantom/static.py", line 112, in from_brainweb
        tissues_mask = get_mri(sub_id, contrast="fuzzy").astype(np.float32)
      File "/opt/hostedtoolcache/Python/3.10.16/x64/lib/python3.10/site-packages/brainweb_dl/mri.py", line 173, in get_mri
        data = _get_mri_sub20(
      File "/opt/hostedtoolcache/Python/3.10.16/x64/lib/python3.10/site-packages/brainweb_dl/mri.py", line 91, in _get_mri_sub20
        filename = get_brainweb20(
      File "/opt/hostedtoolcache/Python/3.10.16/x64/lib/python3.10/site-packages/brainweb_dl/_brainweb.py", line 329, in get_brainweb20
        Parallel(n_jobs=-1, backend="threading")(
      File "/opt/hostedtoolcache/Python/3.10.16/x64/lib/python3.10/site-packages/joblib/parallel.py", line 2007, in __call__
        return output if self.return_generator else list(output)
      File "/opt/hostedtoolcache/Python/3.10.16/x64/lib/python3.10/site-packages/joblib/parallel.py", line 1650, in _get_outputs
        yield from self._retrieve()
      File "/opt/hostedtoolcache/Python/3.10.16/x64/lib/python3.10/site-packages/joblib/parallel.py", line 1754, in _retrieve
        self._raise_error_fast()
      File "/opt/hostedtoolcache/Python/3.10.16/x64/lib/python3.10/site-packages/joblib/parallel.py", line 1789, in _raise_error_fast
        error_job.get_result(self.timeout)
      File "/opt/hostedtoolcache/Python/3.10.16/x64/lib/python3.10/site-packages/joblib/parallel.py", line 745, in get_result
        return self._return_or_raise()
      File "/opt/hostedtoolcache/Python/3.10.16/x64/lib/python3.10/site-packages/joblib/parallel.py", line 763, in _return_or_raise
        raise self._result
    requests.exceptions.ConnectTimeout: HTTPConnectionPool(host='brainweb.bic.mni.mcgill.ca', port=80): Max retries exceeded with url: /cgi/brainweb1/?do_download_alias=subject04_gry&format_value=raw_short&zip_value=gnuzip&download_for_real=%5BStart+download%21%5D (Caused by ConnectTimeoutError(<urllib3.connection.HTTPConnection object at 0x7f2a3221ffd0>, 'Connection to brainweb.bic.mni.mcgill.ca timed out. (connect timeout=None)'))




.. GENERATED FROM PYTHON SOURCE LINES 37-39

Setting up Acquisition Pattern and Initializing Result file.
------------------------------------------------------------

.. GENERATED FROM PYTHON SOURCE LINES 39-51

.. code-block:: Python


    # The next piece of simulation is the acquisition trajectory.
    # Here nothing fancy, we are using a EPI (fully sampled), that samples a 3D
    # k-space (this akin to the 3D EPI sequence of XXXX)

    sampler = EPI3dAcquisitionSampler(accelz=1, acsz=0.1, orderz="top-down")

    smaps = None
    if sim_conf.hardware.n_coils > 1:
        smaps = get_smaps(sim_conf.shape, n_coils=sim_conf.hardware.n_coils)



.. GENERATED FROM PYTHON SOURCE LINES 52-66

Acquisition with Cartesian Engine
---------------------------------

The generated file ``example_EPI.mrd`` does not contains any k-space data for
now, only the sampling trajectory. let's put some in. In order to do so, we
need to setup the **acquisition engine** that models the MR physics, and get
sampled at the specified k-space trajectory.

SNAKE comes with two models for the MR Physics:

- model="simple" :: Each k-space shot acquires a constant signal, which is the
  image contrast at TE.
- model="T2s" :: Each k-space shot is degraded by the T2* decay induced by
  each tissue.

.. GENERATED FROM PYTHON SOURCE LINES 66-105

.. code-block:: Python


    # Here we will use the "simple" model, which is faster.
    #
    # SNAKE's Engine are capable of simulating the data in parallel, by distributing
    # the shots to be acquired to a set of processes. To do so , we need to specify
    # the number of jobs that will run in parallel, as well as the size of a job.
    # Setting the job size and the number of jobs can have a great impact on total
    # runtime and memory consumption.
    #
    # Here, we have a single frame to acquire with 60 frames (one EPI per slice), so
    # a single worker will do.

    from snake.core.engine import EPIAcquisitionEngine

    engine = EPIAcquisitionEngine(model="simple")

    engine("example_EPI.mrd", sampler, phantom, sim_conf, smaps=smaps)

    engine(
        "example_EPI.mrd",
        sampler,
        phantom,
        sim_conf,
        smaps=smaps,
        worker_chunk_size=60,
        n_workers=1,
    )
    engine_t2s = EPIAcquisitionEngine(model="T2s")

    engine_t2s(
        "example_EPI_t2s.mrd",
        sampler,
        phantom,
        sim_conf,
        smaps=smaps,
        worker_chunk_size=60,
        n_workers=1,
    )


.. GENERATED FROM PYTHON SOURCE LINES 106-118

Simple reconstruction
---------------------

Getting k-space data is nice, but
SNAKE also provides rudimentary reconstruction tools to get images (and check
that we didn't mess up the acquisition process).
This is available in the companion package ``snake.toolkit``.

Loading the ``.mrd`` file to retrieve all information can be done using the
``ismrmd`` python package, but SNAKE provides convenient dataloaders, which are
more efficient, and take cares of managing underlying files access. As we are
showcasing the API, we will do things manually here, and use only core SNAKE.

.. GENERATED FROM PYTHON SOURCE LINES 118-141

.. code-block:: Python


    from snake.mrd_utils import CartesianFrameDataLoader
    from scipy.fft import ifftn, ifftshift, fftshift


    def reconstruct_frame(filename):
        with CartesianFrameDataLoader(filename) as data_loader:
            mask, kspace_data = data_loader.get_kspace_frame(0)
        axes = (-3, -2, -1)
        image_data = ifftshift(
            ifftn(fftshift(kspace_data, axes=axes), axes=axes, norm="ortho"), axes=axes
        )

        # Take the square root sum of squares to get the magnitude image (SSOS)
        image_data = np.sqrt(np.sum(np.abs(image_data) ** 2, axis=0))

        return image_data.squeeze().T


    image_simple = reconstruct_frame("example_EPI.mrd")
    image_T2s = reconstruct_frame("example_EPI_t2s.mrd")



.. GENERATED FROM PYTHON SOURCE LINES 142-144

Plotting the result
-------------------

.. GENERATED FROM PYTHON SOURCE LINES 144-160

.. code-block:: Python


    import matplotlib.pyplot as plt
    from snake.toolkit.plotting import axis3dcut

    fig, axs = plt.subplots(1, 3, figsize=(30, 10))

    for ax, img, title in zip(
        axs,
        (image_simple, image_T2s, abs(image_simple - image_T2s)),
        ("simple", "T2s", "diff"),
    ):
        axis3dcut(img, None, None, cbar=True, cuts=(40, 40, 40), width_inches=4, ax=ax)
        ax.set_title(title)

    plt.show()



.. rst-class:: sphx-glr-timing

   **Total running time of the script:** (2 minutes 14.923 seconds)


.. _sphx_glr_download_auto_examples_anatomical_example_T2s_EPI.py:

.. only:: html

  .. container:: sphx-glr-footer sphx-glr-footer-example

    .. container:: sphx-glr-download sphx-glr-download-jupyter

      :download:`Download Jupyter notebook: example_T2s_EPI.ipynb <example_T2s_EPI.ipynb>`

    .. container:: sphx-glr-download sphx-glr-download-python

      :download:`Download Python source code: example_T2s_EPI.py <example_T2s_EPI.py>`

    .. container:: sphx-glr-download sphx-glr-download-zip

      :download:`Download zipped: example_T2s_EPI.zip <example_T2s_EPI.zip>`


.. only:: html

 .. rst-class:: sphx-glr-signature

    `Gallery generated by Sphinx-Gallery <https://sphinx-gallery.github.io>`_
